@@ -1,16 +1,10 @@
 // Copyright 2021 Google LLC
 //
-// Licensed under the Apache License, Version 2.0 (the "License");
-// you may not use this file except in compliance with the License.
-// You may obtain a copy of the License at
-//
-//    https://www.apache.org/licenses/LICENSE-2.0
-//
-// Unless required by applicable law or agreed to in writing, software
-// distributed under the License is distributed on an "AS IS" BASIS,
-// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-// See the License for the specific language governing permissions and
-// limitations under the License.
+// Licensed under the Apache License, Version 2.0 <LICENSE-APACHE or
+// https://www.apache.org/licenses/LICENSE-2.0> or the MIT license
+// <LICENSE-MIT or https://opensource.org/licenses/MIT>, at your
+// option. This file may not be copied, modified, or distributed
+// except according to those terms.
 
 use std::collections::HashSet;
 
@@ -36,10 +30,10 @@ pub(super) struct Discoveries {
 }
 
 impl Discoveries {
-    pub(super) fn search_item(&mut self, item: &Item) {
+    pub(super) fn search_item(&mut self, item: &Item, mod_path: Option<RustPath>) {
         let mut this_mod = PerModDiscoveries {
             discoveries: self,
-            mod_path: None,
+            mod_path,
         };
         this_mod.search_item(item);
     }
@@ -49,6 +43,12 @@ impl Discoveries {
             && self.extern_rust_funs.is_empty()
             && self.extern_rust_types.is_empty()
     }
+
+    pub(crate) fn extend(&mut self, other: Self) {
+        self.cpp_list.extend(other.cpp_list);
+        self.extern_rust_funs.extend(other.extern_rust_funs);
+        self.extern_rust_types.extend(other.extern_rust_types);
+    }
 }
 
 struct PerModDiscoveries<'a> {
@@ -425,7 +425,7 @@ mod tests {
                 }
             }
         };
-        discoveries.search_item(&itm);
+        discoveries.search_item(&itm, None);
         assert_cpp_found(&discoveries);
     }
 
@@ -437,7 +437,7 @@ mod tests {
                 ffi::xxx()
             }
         };
-        discoveries.search_item(&itm);
+        discoveries.search_item(&itm, None);
         assert_cpp_found(&discoveries);
     }
 
@@ -449,7 +449,7 @@ mod tests {
                 ffi::xxx();
             }
         };
-        discoveries.search_item(&itm);
+        discoveries.search_item(&itm, None);
         assert_cpp_found(&discoveries);
     }
 
@@ -461,7 +461,7 @@ mod tests {
                 ffi::a::b::xxx();
             }
         };
-        discoveries.search_item(&itm);
+        discoveries.search_item(&itm, None);
         assert!(!discoveries.cpp_list.is_empty());
         assert!(discoveries.cpp_list.iter().next().unwrap() == "a::b::xxx");
     }
@@ -474,7 +474,7 @@ mod tests {
                 a + 3 * foo(ffi::xxx());
             }
         };
-        discoveries.search_item(&itm);
+        discoveries.search_item(&itm, None);
         assert_cpp_found(&discoveries);
     }
 
@@ -486,7 +486,7 @@ mod tests {
                 let foo: ffi::xxx = bar();
             }
         };
-        discoveries.search_item(&itm);
+        discoveries.search_item(&itm, None);
         assert_cpp_found(&discoveries);
     }
 
@@ -497,7 +497,7 @@ mod tests {
             fn bar(a: &mut ffi::xxx) {
             }
         };
-        discoveries.search_item(&itm);
+        discoveries.search_item(&itm, None);
         assert_cpp_found(&discoveries);
     }
 
@@ -508,7 +508,7 @@ mod tests {
             fn bar(a: cxx::UniquePtr<ffi::xxx>) {
             }
         };
-        discoveries.search_item(&itm);
+        discoveries.search_item(&itm, None);
         assert_cpp_found(&discoveries);
     }
 
@@ -520,17 +520,8 @@ mod tests {
             fn bar(a: cxx::UniquePtr<ffi::xxx>) {
             }
         };
-        discoveries.search_item(&itm);
-        assert!(
-            discoveries
-                .extern_rust_funs
-                .iter()
-                .next()
-                .unwrap()
-                .sig
-                .ident
-                == "bar"
-        );
+        discoveries.search_item(&itm, None);
+        assert!(discoveries.extern_rust_funs.get(0).unwrap().sig.ident == "bar");
     }
 
     #[test]
@@ -542,12 +533,11 @@ mod tests {
 
             }
         };
-        discoveries.search_item(&itm);
+        discoveries.search_item(&itm, None);
         assert!(
             discoveries
                 .extern_rust_types
-                .iter()
-                .next()
+                .get(0)
                 .unwrap()
                 .get_final_ident()
                 == "Bar"
