@@ -1,16 +1,10 @@
 // Copyright 2020 Google LLC
 //
-// Licensed under the Apache License, Version 2.0 (the "License");
-// you may not use this file except in compliance with the License.
-// You may obtain a copy of the License at
-//
-//    https://www.apache.org/licenses/LICENSE-2.0
-//
-// Unless required by applicable law or agreed to in writing, software
-// distributed under the License is distributed on an "AS IS" BASIS,
-// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-// See the License for the specific language governing permissions and
-// limitations under the License.
+// Licensed under the Apache License, Version 2.0 <LICENSE-APACHE or
+// https://www.apache.org/licenses/LICENSE-2.0> or the MIT license
+// <LICENSE-MIT or https://opensource.org/licenses/MIT>, at your
+// option. This file may not be copied, modified, or distributed
+// except according to those terms.
 
 use crate::{
     conversion::api::SubclassName,
@@ -22,15 +16,23 @@ use syn::{parse_quote, Ident, Type};
 pub(crate) enum CppConversionType {
     None,
     FromUniquePtrToValue,
+    FromPtrToValue,
     FromValueToUniquePtr,
+    FromPtrToMove,
 }
 
 impl CppConversionType {
+    /// If we've found a function which does X to its parameter, what
+    /// is the opposite of X? This is used for subclasses where calls
+    /// from Rust to C++ might also involve calls from C++ to Rust.
     fn inverse(&self) -> Self {
         match self {
             CppConversionType::None => CppConversionType::None,
-            CppConversionType::FromUniquePtrToValue => CppConversionType::FromValueToUniquePtr,
+            CppConversionType::FromUniquePtrToValue | CppConversionType::FromPtrToValue => {
+                CppConversionType::FromValueToUniquePtr
+            }
             CppConversionType::FromValueToUniquePtr => CppConversionType::FromUniquePtrToValue,
+            _ => panic!("Did not expect to have to invert this conversion"),
         }
     }
 }
@@ -40,6 +42,19 @@ pub(crate) enum RustConversionType {
     None,
     FromStr,
     ToBoxedUpHolder(SubclassName),
+    FromPinMaybeUninitToPtr,
+    FromPinMoveRefToPtr,
+    FromTypeToPtr,
+    FromValueParamToPtr,
+}
+
+impl RustConversionType {
+    pub(crate) fn requires_mutability(&self) -> Option<syn::token::Mut> {
+        match self {
+            Self::FromPinMoveRefToPtr => Some(parse_quote! { mut }),
+            _ => None,
+        }
+    }
 }
 
 /// A policy for converting types. Conversion may occur on both the Rust and
@@ -77,30 +92,6 @@ impl TypeConversionPolicy {
         }
     }
 
-    pub(crate) fn new_from_unique_ptr(ty: Type) -> Self {
-        TypeConversionPolicy {
-            unwrapped_type: ty,
-            cpp_conversion: CppConversionType::FromUniquePtrToValue,
-            rust_conversion: RustConversionType::None,
-        }
-    }
-
-    pub(crate) fn new_from_str(ty: Type) -> Self {
-        TypeConversionPolicy {
-            unwrapped_type: ty,
-            cpp_conversion: CppConversionType::FromUniquePtrToValue,
-            rust_conversion: RustConversionType::FromStr,
-        }
-    }
-
-    pub(crate) fn box_up_subclass_holder(ty: Type, subclass: SubclassName) -> Self {
-        TypeConversionPolicy {
-            unwrapped_type: ty,
-            cpp_conversion: CppConversionType::None,
-            rust_conversion: RustConversionType::ToBoxedUpHolder(subclass),
-        }
-    }
-
     pub(crate) fn cpp_work_needed(&self) -> bool {
         !matches!(self.cpp_conversion, CppConversionType::None)
     }
@@ -115,6 +106,12 @@ impl TypeConversionPolicy {
     pub(crate) fn converted_rust_type(&self) -> Type {
         match self.cpp_conversion {
             CppConversionType::FromUniquePtrToValue => self.make_unique_ptr_type(),
+            CppConversionType::FromPtrToValue => {
+                let innerty = &self.unwrapped_type;
+                parse_quote! {
+                    *mut #innerty
+                }
+            }
             _ => self.unwrapped_type.clone(),
         }
     }
@@ -130,6 +127,9 @@ impl TypeConversionPolicy {
         !matches!(self.rust_conversion, RustConversionType::None)
     }
 
+    /// Subclass support involves calls from Rust -> C++, but
+    /// also from C++ -> Rust. Work out the correct argument conversion
+    /// type for the latter call, when given the former.
     pub(crate) fn inverse(&self) -> Self {
         Self {
             unwrapped_type: self.unwrapped_type.clone(),
@@ -137,30 +137,42 @@ impl TypeConversionPolicy {
             rust_conversion: self.rust_conversion.clone(),
         }
     }
+
+    pub(crate) fn bridge_unsafe_needed(&self) -> bool {
+        matches!(
+            self.rust_conversion,
+            RustConversionType::FromValueParamToPtr
+        )
+    }
 }
 
 #[derive(Clone)]
-
 pub(crate) enum CppFunctionBody {
     FunctionCall(Namespace, Ident),
     StaticMethodCall(Namespace, Ident, Ident),
-    Constructor,
+    PlacementNew(Namespace, Ident),
+    MakeUnique,
     ConstructSuperclass(String),
+    Cast,
+    Destructor(Namespace, Ident),
+    AllocUninitialized(QualifiedName),
+    FreeUninitialized(QualifiedName),
 }
 
 #[derive(Clone)]
-
 pub(crate) enum CppFunctionKind {
     Function,
     Method,
-    ConstMethod,
     Constructor,
+    ConstMethod,
+    SynthesizedConstructor,
 }
 
 #[derive(Clone)]
 pub(crate) struct CppFunction {
     pub(crate) payload: CppFunctionBody,
     pub(crate) wrapper_function_name: Ident,
+    pub(crate) original_cpp_name: String,
     pub(crate) return_conversion: Option<TypeConversionPolicy>,
     pub(crate) argument_conversion: Vec<TypeConversionPolicy>,
     pub(crate) kind: CppFunctionKind,
