@@ -1,81 +1,252 @@
 // Copyright 2020 Google LLC
 //
-// Licensed under the Apache License, Version 2.0 (the "License");
-// you may not use this file except in compliance with the License.
-// You may obtain a copy of the License at
-//
-//    https://www.apache.org/licenses/LICENSE-2.0
-//
-// Unless required by applicable law or agreed to in writing, software
-// distributed under the License is distributed on an "AS IS" BASIS,
-// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-// See the License for the specific language governing permissions and
-// limitations under the License.
+// Licensed under the Apache License, Version 2.0 <LICENSE-APACHE or
+// https://www.apache.org/licenses/LICENSE-2.0> or the MIT license
+// <LICENSE-MIT or https://opensource.org/licenses/MIT>, at your
+// option. This file may not be copied, modified, or distributed
+// except according to those terms.
 
 use std::collections::HashSet;
 
 use crate::types::{make_ident, Namespace, QualifiedName};
 use autocxx_parser::RustPath;
+use itertools::Itertools;
+use quote::ToTokens;
 use syn::{
-    punctuated::Punctuated, token::Comma, Attribute, FnArg, Ident, ImplItem, ItemConst, ItemEnum,
-    ItemStruct, ItemType, ItemUse, ReturnType, Signature, Type, Visibility,
+    parse::Parse,
+    punctuated::Punctuated,
+    token::{Comma, Unsafe},
+    Attribute, FnArg, Ident, ItemConst, ItemEnum, ItemStruct, ItemType, ItemUse, LitBool, LitInt,
+    Pat, ReturnType, Signature, Type, Visibility,
 };
 
 use super::{
-    analysis::fun::{function_wrapper::CppFunction, ReceiverMutability},
+    analysis::fun::{
+        function_wrapper::{CppFunction, CppFunctionBody, CppFunctionKind},
+        ReceiverMutability,
+    },
     convert_error::{ConvertErrorWithContext, ErrorContext},
     ConvertError,
 };
 
 #[derive(Copy, Clone, Eq, PartialEq)]
 pub(crate) enum TypeKind {
-    Pod,          // trivial. Can be moved and copied in Rust.
-    NonPod,       // has destructor or non-trivial move constructors. Can only hold by UniquePtr
-    NonPodNested, // same, but nested inside another C++ struct/class.
-    // We have to do different codegen here because cxx can't cope with
-    // nested classes declared as 'type X;' so we instead have to do
-    // 'type X = super::bindgen::X;'
+    Pod,    // trivial. Can be moved and copied in Rust.
+    NonPod, // has destructor or non-trivial move constructors. Can only hold by UniquePtr
     Abstract, // has pure virtual members - can't even generate UniquePtr.
-              // It's possible that the type itself isn't pure virtual, but it inherits from
-              // some other type which is pure virtual. Alternatively, maybe we just don't
-              // know if the base class is pure virtual because it wasn't on the allowlist,
-              // in which case we'll err on the side of caution.
-}
-
-/// An entry which needs to go into an `impl` block for a given type.
-pub(crate) struct ImplBlockDetails {
-    pub(crate) item: ImplItem,
-    pub(crate) ty: Ident,
+            // It's possible that the type itself isn't pure virtual, but it inherits from
+            // some other type which is pure virtual. Alternatively, maybe we just don't
+            // know if the base class is pure virtual because it wasn't on the allowlist,
+            // in which case we'll err on the side of caution.
 }
 
 /// C++ visibility.
-#[derive(Clone, PartialEq, Eq, Copy)]
+#[derive(Debug, Clone, PartialEq, Eq, Copy)]
 pub(crate) enum CppVisibility {
     Public,
     Protected,
     Private,
 }
 
+/// Details about a C++ struct.
+pub(crate) struct StructDetails {
+    pub(crate) vis: CppVisibility,
+    pub(crate) item: ItemStruct,
+    pub(crate) layout: Option<Layout>,
+    pub(crate) has_rvalue_reference_fields: bool,
+}
+
+/// Layout of a type, equivalent to the same type in ir/layout.rs in bindgen
+#[derive(Clone)]
+pub(crate) struct Layout {
+    /// The size (in bytes) of this layout.
+    pub(crate) size: usize,
+    /// The alignment (in bytes) of this layout.
+    pub(crate) align: usize,
+    /// Whether this layout's members are packed or not.
+    pub(crate) packed: bool,
+}
+
+impl Parse for Layout {
+    fn parse(input: syn::parse::ParseStream) -> syn::Result<Self> {
+        let size: LitInt = input.parse()?;
+        input.parse::<syn::token::Comma>()?;
+        let align: LitInt = input.parse()?;
+        input.parse::<syn::token::Comma>()?;
+        let packed: LitBool = input.parse()?;
+        Ok(Layout {
+            size: size.base10_parse().unwrap(),
+            align: align.base10_parse().unwrap(),
+            packed: packed.value(),
+        })
+    }
+}
+
+#[derive(Clone)]
+pub(crate) enum Virtualness {
+    None,
+    Virtual,
+    PureVirtual,
+}
+
+#[derive(Clone, Copy)]
+pub(crate) enum CastMutability {
+    ConstToConst,
+    MutToConst,
+    MutToMut,
+}
+
+/// Indicates that this function (which is synthetic) should
+/// be a trait implementation rather than a method or free function.
+#[derive(Clone)]
+pub(crate) enum TraitSynthesis {
+    Cast {
+        to_type: QualifiedName,
+        mutable: CastMutability,
+    },
+    AllocUninitialized(QualifiedName),
+    FreeUninitialized(QualifiedName),
+}
+
+/// Details of a subclass constructor.
+/// TODO: zap this; replace with an extra API.
+#[derive(Clone)]
+pub(crate) struct SubclassConstructorDetails {
+    pub(crate) subclass: SubclassName,
+    pub(crate) is_trivial: bool,
+    /// Implementation of the constructor _itself_ as distinct
+    /// from any wrapper function we create to call it.
+    pub(crate) cpp_impl: CppFunction,
+}
+
+/// Contributions to traits representing C++ superclasses that
+/// we may implement as Rust subclasses.
+#[derive(Clone)]
+pub(crate) struct SuperclassMethod {
+    pub(crate) name: Ident,
+    pub(crate) receiver: QualifiedName,
+    pub(crate) params: Punctuated<FnArg, Comma>,
+    pub(crate) param_names: Vec<Pat>,
+    pub(crate) ret_type: ReturnType,
+    pub(crate) receiver_mutability: ReceiverMutability,
+    pub(crate) requires_unsafe: UnsafetyNeeded,
+    pub(crate) is_pure_virtual: bool,
+}
+
+/// Information about references (as opposed to pointers) to be found
+/// within the function signature. This is derived from bindgen annotations
+/// which is why it's not within `FuncToConvert::inputs`
+#[derive(Default, Clone)]
+pub(crate) struct References {
+    pub(crate) rvalue_ref_params: HashSet<Ident>,
+    pub(crate) ref_params: HashSet<Ident>,
+    pub(crate) ref_return: bool,
+    pub(crate) rvalue_ref_return: bool,
+}
+
+impl References {
+    pub(crate) fn new_with_this_and_return_as_reference() -> Self {
+        Self {
+            ref_return: true,
+            ref_params: [make_ident("this")].into_iter().collect(),
+            ..Default::default()
+        }
+    }
+}
+
+#[derive(Clone)]
+pub(crate) struct TraitImplSignature {
+    pub(crate) ty: Type,
+    pub(crate) trait_signature: Type,
+    /// The trait is 'unsafe' itself
+    pub(crate) unsafety: Option<Unsafe>,
+}
+
+impl Eq for TraitImplSignature {}
+
+impl PartialEq for TraitImplSignature {
+    fn eq(&self, other: &Self) -> bool {
+        totokens_equal(&self.unsafety, &other.unsafety)
+            && totokens_equal(&self.ty, &other.ty)
+            && totokens_equal(&self.trait_signature, &other.trait_signature)
+    }
+}
+
+fn totokens_to_string<T: ToTokens>(a: &T) -> String {
+    a.to_token_stream().to_string()
+}
+
+fn totokens_equal<T: ToTokens>(a: &T, b: &T) -> bool {
+    totokens_to_string(a) == totokens_to_string(b)
+}
+
+fn hash_totokens<T: ToTokens, H: std::hash::Hasher>(a: &T, state: &mut H) {
+    use std::hash::Hash;
+    totokens_to_string(a).hash(state)
+}
+
+impl std::hash::Hash for TraitImplSignature {
+    fn hash<H: std::hash::Hasher>(&self, state: &mut H) {
+        hash_totokens(&self.ty, state);
+        hash_totokens(&self.trait_signature, state);
+        hash_totokens(&self.unsafety, state);
+    }
+}
+
+#[derive(Clone, Debug)]
+pub(crate) enum SpecialMemberKind {
+    DefaultConstructor,
+    CopyConstructor,
+    MoveConstructor,
+    Destructor,
+    AssignmentOperator,
+}
+
+#[derive(Clone)]
+pub(crate) enum Provenance {
+    Bindgen,
+    SynthesizedOther,
+    SynthesizedMakeUnique,
+    SynthesizedSubclassConstructor(Box<SubclassConstructorDetails>),
+}
+
 /// A C++ function for which we need to generate bindings, but haven't
 /// yet analyzed in depth. This is little more than a `ForeignItemFn`
 /// broken down into its constituent parts, plus some metadata from the
 /// surrounding bindgen parsing context.
+///
+/// Some parts of the code synthesize additional functions and then
+/// pass them through the same pipeline _as if_ they were discovered
+/// during normal bindgen parsing. If that happens, they'll create one
+/// of these structures, and typically fill in some of the
+/// `synthesized_*` members which are not filled in from bindgen.
 #[derive(Clone)]
 pub(crate) struct FuncToConvert {
+    pub(crate) provenance: Provenance,
     pub(crate) ident: Ident,
     pub(crate) doc_attr: Option<Attribute>,
     pub(crate) inputs: Punctuated<FnArg, Comma>,
     pub(crate) output: ReturnType,
     pub(crate) vis: Visibility,
-    pub(crate) is_pure_virtual: bool,
+    pub(crate) virtualness: Virtualness,
     pub(crate) cpp_vis: CppVisibility,
-    pub(crate) is_move_constructor: bool,
+    pub(crate) special_member: Option<SpecialMemberKind>,
     pub(crate) unused_template_param: bool,
-    pub(crate) return_type_is_reference: bool,
-    pub(crate) reference_args: HashSet<Ident>,
+    pub(crate) references: References,
     pub(crate) original_name: Option<String>,
-    pub(crate) virtual_this_type: Option<QualifiedName>,
+    /// Used for static functions only. For all other functons,
+    /// this is figured out from the receiver type in the inputs.
     pub(crate) self_ty: Option<QualifiedName>,
+    /// If we wish to use a different 'this' type than the original
+    /// method receiver, e.g. because we're making a subclass
+    /// constructor, fill it in here.
+    pub(crate) synthesized_this_type: Option<QualifiedName>,
+    /// If this function should actually belong to a trait.
+    pub(crate) add_to_trait: Option<TraitSynthesis>,
+    /// If Some, this function didn't really exist in the original
+    /// C++ and instead we're synthesizing it.
+    pub(crate) synthetic_cpp: Option<(CppFunctionBody, CppFunctionKind)>,
+    pub(crate) is_deleted: bool,
 }
 
 /// Layers of analysis which may be applied to decorate each API.
@@ -106,7 +277,7 @@ pub(crate) enum TypedefKind {
 #[derive(Clone, Hash, PartialEq, Eq)]
 pub(crate) struct ApiName {
     pub(crate) name: QualifiedName,
-    pub(crate) cpp_name: Option<String>,
+    cpp_name: Option<String>,
 }
 
 impl ApiName {
@@ -132,11 +303,24 @@ impl ApiName {
         Self::new(&Namespace::new(), id)
     }
 
-    pub(crate) fn is_nested_struct_or_class(&self) -> bool {
+    pub(crate) fn cpp_name(&self) -> String {
         self.cpp_name
             .as_ref()
-            .map(|n| n.contains("::"))
-            .unwrap_or_default()
+            .cloned()
+            .unwrap_or_else(|| self.name.get_final_item().to_string())
+    }
+
+    pub(crate) fn qualified_cpp_name(&self) -> String {
+        let cpp_name = self.cpp_name();
+        self.name
+            .ns_segment_iter()
+            .cloned()
+            .chain(std::iter::once(cpp_name))
+            .join("::")
+    }
+
+    pub(crate) fn cpp_name_if_present(&self) -> Option<&String> {
+        self.cpp_name.as_ref()
     }
 }
 
@@ -185,6 +369,17 @@ impl SubclassName {
     fn with_suffix(&self, suffix: &str) -> Ident {
         make_ident(format!("{}{}", self.0.name.get_final_item(), suffix))
     }
+    pub(crate) fn get_trait_api_name(sup: &QualifiedName, method_name: &str) -> QualifiedName {
+        QualifiedName::new(
+            sup.get_namespace(),
+            make_ident(format!(
+                "{}_{}_trait_item",
+                sup.get_final_item(),
+                method_name
+            )),
+        )
+    }
+    // TODO this and the following should probably include both class name and method name
     pub(crate) fn get_super_fn_name(superclass_namespace: &Namespace, id: &str) -> QualifiedName {
         let id = make_ident(format!("{}_super", id));
         QualifiedName::new(superclass_namespace, id)
@@ -258,7 +453,7 @@ pub(crate) enum Api<T: AnalysisPhase> {
     /// `bindgen` output.
     Struct {
         name: ApiName,
-        item: ItemStruct,
+        details: Box<StructDetails>,
         analysis: T::StructAnalysis,
     },
     /// A variable-length C integer type (e.g. int, unsigned long).
@@ -289,18 +484,17 @@ pub(crate) enum Api<T: AnalysisPhase> {
         subclass: SubclassName,
         details: Box<RustSubclassFnDetails>,
     },
-    // A constructor for a subclass.
-    RustSubclassConstructor {
-        name: ApiName,
-        subclass: SubclassName,
-        cpp_impl: Box<CppFunction>,
-        is_trivial: bool,
-    },
     /// A Rust subclass of a C++ class.
     Subclass {
         name: SubclassName,
         superclass: QualifiedName,
     },
+    /// Contributions to the traits representing superclass methods that we might
+    /// subclass in Rust.
+    SubclassTraitItem {
+        name: ApiName,
+        details: SuperclassMethod,
+    },
 }
 
 pub(crate) struct RustSubclassFnDetails {
@@ -310,12 +504,20 @@ pub(crate) struct RustSubclassFnDetails {
     pub(crate) method_name: Ident,
     pub(crate) superclass: QualifiedName,
     pub(crate) receiver_mutability: ReceiverMutability,
-    pub(crate) dependency: QualifiedName,
-    pub(crate) requires_unsafe: bool,
+    pub(crate) dependencies: Vec<QualifiedName>,
+    pub(crate) requires_unsafe: UnsafetyNeeded,
+    pub(crate) is_pure_virtual: bool,
+}
+
+#[derive(Clone, Debug)]
+pub(crate) enum UnsafetyNeeded {
+    None,
+    JustBridge,
+    Always,
 }
 
 impl<T: AnalysisPhase> Api<T> {
-    fn name_info(&self) -> &ApiName {
+    pub(crate) fn name_info(&self) -> &ApiName {
         match self {
             Api::ForwardDeclaration { name } => name,
             Api::ConcreteType { name, .. } => name,
@@ -330,8 +532,8 @@ impl<T: AnalysisPhase> Api<T> {
             Api::RustType { name, .. } => name,
             Api::RustFn { name, .. } => name,
             Api::RustSubclassFn { name, .. } => name,
-            Api::RustSubclassConstructor { name, .. } => name,
             Api::Subclass { name, .. } => &name.0,
+            Api::SubclassTraitItem { name, .. } => name,
         }
     }
 
@@ -359,12 +561,19 @@ impl<T: AnalysisPhase> Api<T> {
             .unwrap_or_else(|| self.name().get_final_item())
     }
 
+    /// If this API turns out to have the same QualifiedName as another,
+    /// whether it's OK to just discard it?
+    pub(crate) fn discard_duplicates(&self) -> bool {
+        matches!(self, Api::IgnoredItem { .. })
+    }
+
     pub(crate) fn valid_types(&self) -> Box<dyn Iterator<Item = QualifiedName>> {
         match self {
             Api::Subclass { name, .. } => Box::new(
                 vec![
                     self.name().clone(),
                     QualifiedName::new(&Namespace::new(), name.holder()),
+                    name.cpp(),
                 ]
                 .into_iter(),
             ),
@@ -401,7 +610,7 @@ impl<T: AnalysisPhase> Api<T> {
 
     pub(crate) fn struct_unchanged(
         name: ApiName,
-        item: ItemStruct,
+        details: Box<StructDetails>,
         analysis: T::StructAnalysis,
     ) -> Result<Box<dyn Iterator<Item = Api<T>>>, ConvertErrorWithContext>
     where
@@ -409,7 +618,7 @@ impl<T: AnalysisPhase> Api<T> {
     {
         Ok(Box::new(std::iter::once(Api::Struct {
             name,
-            item,
+            details,
             analysis,
         })))
     }
