@@ -1,22 +1,15 @@
 // Copyright 2020 Google LLC
 //
-// Licensed under the Apache License, Version 2.0 (the "License");
-// you may not use this file except in compliance with the License.
-// You may obtain a copy of the License at
-//
-//    https://www.apache.org/licenses/LICENSE-2.0
-//
-// Unless required by applicable law or agreed to in writing, software
-// distributed under the License is distributed on an "AS IS" BASIS,
-// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-// See the License for the specific language governing permissions and
-// limitations under the License.
+// Licensed under the Apache License, Version 2.0 <LICENSE-APACHE or
+// https://www.apache.org/licenses/LICENSE-2.0> or the MIT license
+// <LICENSE-MIT or https://opensource.org/licenses/MIT>, at your
+// option. This file may not be copied, modified, or distributed
+// except according to those terms.
 
 use crate::conversion::{
     analysis::fun::function_wrapper::{CppConversionType, TypeConversionPolicy},
     ConvertError,
 };
-use crate::known_types::type_lacks_copy_constructor;
 
 use super::type_to_cpp::{type_to_cpp, CppNameMap};
 
@@ -26,14 +19,17 @@ impl TypeConversionPolicy {
         cpp_name_map: &CppNameMap,
     ) -> Result<String, ConvertError> {
         match self.cpp_conversion {
-            CppConversionType::FromUniquePtrToValue => self.wrapped_type(cpp_name_map),
+            CppConversionType::FromUniquePtrToValue => self.unique_ptr_wrapped_type(cpp_name_map),
+            CppConversionType::FromPtrToValue => {
+                Ok(format!("{}*", self.unwrapped_type_as_string(cpp_name_map)?))
+            }
             _ => self.unwrapped_type_as_string(cpp_name_map),
         }
     }
 
     pub(super) fn converted_type(&self, cpp_name_map: &CppNameMap) -> Result<String, ConvertError> {
         match self.cpp_conversion {
-            CppConversionType::FromValueToUniquePtr => self.wrapped_type(cpp_name_map),
+            CppConversionType::FromValueToUniquePtr => self.unique_ptr_wrapped_type(cpp_name_map),
             _ => self.unwrapped_type_as_string(cpp_name_map),
         }
     }
@@ -42,7 +38,10 @@ impl TypeConversionPolicy {
         type_to_cpp(&self.unwrapped_type, cpp_name_map)
     }
 
-    fn wrapped_type(&self, original_name_map: &CppNameMap) -> Result<String, ConvertError> {
+    fn unique_ptr_wrapped_type(
+        &self,
+        original_name_map: &CppNameMap,
+    ) -> Result<String, ConvertError> {
         Ok(format!(
             "std::unique_ptr<{}>",
             self.unwrapped_type_as_string(original_name_map)?
@@ -53,22 +52,34 @@ impl TypeConversionPolicy {
         &self,
         var_name: &str,
         cpp_name_map: &CppNameMap,
-        use_rvo: bool,
+        is_return: bool,
     ) -> Result<String, ConvertError> {
+        // If is_return we want to avoid unnecessary std::moves because they
+        // make RVO less effective
         Ok(match self.cpp_conversion {
             CppConversionType::None => {
-                if type_lacks_copy_constructor(&self.unwrapped_type) && !use_rvo {
-                    format!("std::move({})", var_name)
-                } else {
+                if is_return {
                     var_name.to_string()
+                } else {
+                    format!("std::move({})", var_name)
                 }
             }
-            CppConversionType::FromUniquePtrToValue => format!("std::move(*{})", var_name),
+            CppConversionType::FromUniquePtrToValue | CppConversionType::FromPtrToMove => {
+                format!("std::move(*{})", var_name)
+            }
             CppConversionType::FromValueToUniquePtr => format!(
                 "std::make_unique<{}>({})",
                 self.unconverted_type(cpp_name_map)?,
                 var_name
             ),
+            CppConversionType::FromPtrToValue => {
+                let dereference = format!("*{}", var_name);
+                if is_return {
+                    dereference
+                } else {
+                    format!("std::move({})", dereference)
+                }
+            }
         })
     }
 }
