@@ -1,16 +1,10 @@
 // Copyright 2020 Google LLC
 //
-// Licensed under the Apache License, Version 2.0 (the "License");
-// you may not use this file except in compliance with the License.
-// You may obtain a copy of the License at
-//
-//    https://www.apache.org/licenses/LICENSE-2.0
-//
-// Unless required by applicable law or agreed to in writing, software
-// distributed under the License is distributed on an "AS IS" BASIS,
-// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-// See the License for the specific language governing permissions and
-// limitations under the License.
+// Licensed under the Apache License, Version 2.0 <LICENSE-APACHE or
+// https://www.apache.org/licenses/LICENSE-2.0> or the MIT license
+// <LICENSE-MIT or https://opensource.org/licenses/MIT>, at your
+// option. This file may not be copied, modified, or distributed
+// except according to those terms.
 
 mod fun_codegen;
 mod function_wrapper_rs;
@@ -24,15 +18,15 @@ use std::collections::{HashMap, HashSet};
 
 use autocxx_parser::IncludeCppConfig;
 
+use itertools::Itertools;
 use proc_macro2::{Span, TokenStream};
 use syn::{
     parse_quote, punctuated::Punctuated, token::Comma, Attribute, Expr, FnArg, ForeignItem,
-    ForeignItemFn, Ident, ImplItem, Item, ItemForeignMod, ItemMod, Pat, ReturnType, TraitItem,
+    ForeignItemFn, Ident, ImplItem, Item, ItemForeignMod, ItemMod, TraitItem,
 };
 
 use crate::{
     conversion::{
-        analysis::fun::MethodKind,
         codegen_rs::{
             non_pod_struct::{make_non_pod, new_non_pod_struct},
             unqualify::{unqualify_params, unqualify_ret_type},
@@ -50,20 +44,29 @@ use self::{
 };
 
 use super::{
-    analysis::fun::{FnAnalysis, FnKind},
-    api::RustSubclassFnDetails,
+    analysis::fun::{FnPhase, ReceiverMutability},
+    api::{AnalysisPhase, Api, SubclassName, TypeKind, TypedefKind},
+};
+use super::{
+    api::{Layout, Provenance, RustSubclassFnDetails, SuperclassMethod, TraitImplSignature},
+    apivec::ApiVec,
     codegen_cpp::type_to_cpp::{
         namespaced_name_using_original_name_map, original_name_map_from_apis, CppNameMap,
     },
 };
-use super::{
-    analysis::fun::{FnPhase, ReceiverMutability},
-    api::{AnalysisPhase, Api, ImplBlockDetails, SubclassName, TypeKind, TypedefKind},
-};
 use super::{convert_error::ErrorContext, ConvertError};
 use quote::{quote, ToTokens};
 
-unzip_n::unzip_n!(pub 4);
+/// An entry which needs to go into an `impl` block for a given type.
+struct ImplBlockDetails {
+    item: ImplItem,
+    ty: Ident,
+}
+
+struct TraitImplBlockDetails {
+    item: TraitItem,
+    key: TraitImplSignature,
+}
 
 /// Whether and how this item should be exposed in the mods constructed
 /// for actual end-user use.
@@ -123,15 +126,6 @@ fn get_string_items() -> Vec<Item> {
     .to_vec()
 }
 
-struct SuperclassMethod {
-    name: Ident,
-    params: Punctuated<FnArg, Comma>,
-    param_names: Vec<Pat>,
-    ret_type: ReturnType,
-    receiver_mutability: ReceiverMutability,
-    requires_unsafe: bool,
-}
-
 /// Type which handles generation of Rust code.
 /// In practice, much of the "generation" involves connecting together
 /// existing lumps of code within the Api structures.
@@ -140,26 +134,29 @@ pub(crate) struct RsCodeGenerator<'a> {
     bindgen_mod: ItemMod,
     original_name_map: CppNameMap,
     config: &'a IncludeCppConfig,
+    header_name: Option<String>,
 }
 
 impl<'a> RsCodeGenerator<'a> {
     /// Generate code for a set of APIs that was discovered during parsing.
     pub(crate) fn generate_rs_code(
-        all_apis: Vec<Api<FnPhase>>,
+        all_apis: ApiVec<FnPhase>,
         include_list: &'a [String],
         bindgen_mod: ItemMod,
         config: &'a IncludeCppConfig,
+        header_name: Option<String>,
     ) -> Vec<Item> {
         let c = Self {
             include_list,
             bindgen_mod,
             original_name_map: original_name_map_from_apis(&all_apis),
             config,
+            header_name,
         };
         c.rs_codegen(all_apis)
     }
 
-    fn rs_codegen(mut self, all_apis: Vec<Api<FnPhase>>) -> Vec<Item> {
+    fn rs_codegen(mut self, all_apis: ApiVec<FnPhase>) -> Vec<Item> {
         // ... and now let's start to generate the output code.
         // First off, when we generate structs we may need to add some methods
         // if they're superclasses.
@@ -191,18 +188,22 @@ impl<'a> RsCodeGenerator<'a> {
         // sub-mods by namespace. From here on, things are flat.
         let (_, rs_codegen_results): (Vec<_>, Vec<_>) =
             rs_codegen_results_and_namespaces.into_iter().unzip();
-        let (extern_c_mod_items, extern_rust_mod_items, all_items, bridge_items) =
-            rs_codegen_results
-                .into_iter()
-                .map(|api| {
-                    (
-                        api.extern_c_mod_items,
-                        api.extern_rust_mod_items,
-                        api.global_items,
-                        api.bridge_items,
-                    )
-                })
-                .unzip_n_vec();
+        let (extern_c_mod_items, extern_rust_mod_items, all_items, bridge_items): (
+            Vec<_>,
+            Vec<_>,
+            Vec<_>,
+            Vec<_>,
+        ) = rs_codegen_results
+            .into_iter()
+            .map(|api| {
+                (
+                    api.extern_c_mod_items,
+                    api.extern_rust_mod_items,
+                    api.global_items,
+                    api.bridge_items,
+                )
+            })
+            .multiunzip();
         // Items for the [cxx::bridge] mod...
         let mut bridge_items: Vec<Item> = bridge_items.into_iter().flatten().collect();
         // Things to include in the "extern "C"" mod passed within the cxx::bridge
@@ -258,7 +259,7 @@ impl<'a> RsCodeGenerator<'a> {
 
     fn accumulate_superclass_methods(
         &self,
-        apis: &[Api<FnPhase>],
+        apis: &ApiVec<FnPhase>,
     ) -> HashMap<QualifiedName, Vec<SuperclassMethod>> {
         let mut results = HashMap::new();
         results.extend(
@@ -266,38 +267,11 @@ impl<'a> RsCodeGenerator<'a> {
                 .superclasses()
                 .map(|sc| (QualifiedName::new_from_cpp_name(sc), Vec::new())),
         );
-        for api in apis {
-            if let Api::Function {
-                name,
-                analysis:
-                    FnAnalysis {
-                        kind: FnKind::Method(receiver, method_kind),
-                        params,
-                        ret_type,
-                        param_details,
-                        ..
-                    },
-                ..
-            } = api
-            {
-                match method_kind {
-                    MethodKind::Virtual(receiver_mutability)
-                    | MethodKind::PureVirtual(receiver_mutability) => {
-                        let list = results.get_mut(receiver);
-                        if let Some(list) = list {
-                            let param_names =
-                                param_details.iter().map(|pd| pd.name.clone()).collect();
-                            list.push(SuperclassMethod {
-                                name: name.name.get_final_ident(),
-                                params: params.clone(),
-                                ret_type: ret_type.clone(),
-                                param_names,
-                                receiver_mutability: receiver_mutability.clone(),
-                                requires_unsafe: param_details.iter().any(|pd| pd.requires_unsafe),
-                            })
-                        }
-                    }
-                    _ => {} // otherwise, this is not a superclass of anything specified as a subclass!(..)
+        for api in apis.iter() {
+            if let Api::SubclassTraitItem { details, .. } = api {
+                let list = results.get_mut(&details.receiver);
+                if let Some(list) = list {
+                    list.push(details.clone());
                 }
             }
         }
@@ -314,10 +288,7 @@ impl<'a> RsCodeGenerator<'a> {
 
     fn build_include_foreign_items(&self, has_additional_cpp_needs: bool) -> Vec<ForeignItem> {
         let extra_inclusion = if has_additional_cpp_needs {
-            Some(format!(
-                "autocxxgen_{}.h",
-                self.config.get_mod_name().to_string()
-            ))
+            Some(self.header_name.clone().unwrap())
         } else {
             None
         };
@@ -413,6 +384,7 @@ impl<'a> RsCodeGenerator<'a> {
         ns: &Namespace,
     ) {
         let mut impl_entries_by_type: HashMap<_, Vec<_>> = HashMap::new();
+        let mut trait_impl_entries_by_trait_and_ty: HashMap<_, Vec<_>> = HashMap::new();
         for item in ns_entries.entries() {
             output_items.extend(item.1.bindgen_mod_items.iter().cloned());
             if let Some(impl_entry) = &item.1.impl_entry {
@@ -421,6 +393,12 @@ impl<'a> RsCodeGenerator<'a> {
                     .or_default()
                     .push(&impl_entry.item);
             }
+            if let Some(trait_impl_entry) = &item.1.trait_impl_entry {
+                trait_impl_entries_by_trait_and_ty
+                    .entry(trait_impl_entry.key.clone())
+                    .or_default()
+                    .push(&trait_impl_entry.item);
+            }
         }
         for (ty, entries) in impl_entries_by_type.into_iter() {
             output_items.push(Item::Impl(parse_quote! {
@@ -429,6 +407,16 @@ impl<'a> RsCodeGenerator<'a> {
                 }
             }))
         }
+        for (key, entries) in trait_impl_entries_by_trait_and_ty.into_iter() {
+            let unsafety = key.unsafety;
+            let ty = key.ty;
+            let trt = key.trait_signature;
+            output_items.push(Item::Impl(parse_quote! {
+                #unsafety impl #trt for #ty {
+                    #(#entries)*
+                }
+            }))
+        }
         for (child_name, child_ns_entries) in ns_entries.children() {
             let new_ns = ns.push((*child_name).clone());
             let child_id = make_ident(child_name);
@@ -479,48 +467,43 @@ impl<'a> RsCodeGenerator<'a> {
                     extern_c_mod_items: vec![ForeignItem::Fn(parse_quote!(
                         fn #make_string_name(str_: &str) -> UniquePtr<CxxString>;
                     ))],
-                    bridge_items: Vec::new(),
                     global_items: get_string_items(),
-                    bindgen_mod_items: Vec::new(),
-                    impl_entry: None,
                     materializations: vec![Use::UsedFromCxxBridgeWithAlias(make_ident(
                         "make_string",
                     ))],
-                    extern_rust_mod_items: Vec::new(),
+                    ..Default::default()
                 }
             }
             Api::Function { fun, analysis, .. } => {
                 gen_function(name.get_namespace(), *fun, analysis, cpp_call_name)
             }
             Api::Const { const_item, .. } => RsCodegenResult {
-                global_items: Vec::new(),
-                impl_entry: None,
-                bridge_items: Vec::new(),
-                extern_c_mod_items: Vec::new(),
                 bindgen_mod_items: vec![Item::Const(const_item)],
                 materializations: vec![Use::UsedFromBindgen],
-                extern_rust_mod_items: Vec::new(),
+                ..Default::default()
             },
             Api::Typedef { analysis, .. } => RsCodegenResult {
-                extern_c_mod_items: Vec::new(),
-                bridge_items: Vec::new(),
-                global_items: Vec::new(),
                 bindgen_mod_items: vec![match analysis.kind {
                     TypedefKind::Type(type_item) => Item::Type(type_item),
                     TypedefKind::Use(use_item) => Item::Use(use_item),
                 }],
-                impl_entry: None,
                 materializations: vec![Use::UsedFromBindgen],
-                extern_rust_mod_items: Vec::new(),
+                ..Default::default()
             },
-            Api::Struct { item, analysis, .. } => {
-                let doc_attr = get_doc_attr(&item.attrs);
+            Api::Struct {
+                details, analysis, ..
+            } => {
+                let doc_attr = get_doc_attr(&details.item.attrs);
+                let layout = details.layout.clone();
                 self.generate_type(
                     &name,
                     id,
-                    analysis.kind,
-                    || Some((Item::Struct(item), doc_attr)),
+                    analysis.pod.kind,
+                    analysis.constructors.move_constructor,
+                    analysis.constructors.destructor,
+                    || Some((Item::Struct(details.item), doc_attr)),
                     associated_methods,
+                    layout,
                 )
             }
             Api::Enum { item, .. } => {
@@ -529,49 +512,46 @@ impl<'a> RsCodeGenerator<'a> {
                     &name,
                     id,
                     TypeKind::Pod,
+                    true,
+                    true,
                     || Some((Item::Enum(item), doc_attr)),
                     associated_methods,
+                    None,
                 )
             }
-            Api::ForwardDeclaration { .. } | Api::ConcreteType { .. } => {
-                self.generate_type(&name, id, TypeKind::Abstract, || None, associated_methods)
-            }
+            Api::ForwardDeclaration { .. } | Api::ConcreteType { .. } => self.generate_type(
+                &name,
+                id,
+                TypeKind::Abstract,
+                false, // assume for now that these types can't be kept in a Vector
+                true,  // assume for now that these types can be put in a smart pointer
+                || None,
+                associated_methods,
+                None,
+            ),
             Api::CType { .. } => RsCodegenResult {
-                global_items: Vec::new(),
-                impl_entry: None,
-                bridge_items: Vec::new(),
                 extern_c_mod_items: vec![ForeignItem::Verbatim(quote! {
                     type #id = autocxx::#id;
                 })],
-                bindgen_mod_items: Vec::new(),
-                materializations: Vec::new(),
-                extern_rust_mod_items: Vec::new(),
+                ..Default::default()
             },
             Api::RustType { path, .. } => RsCodegenResult {
-                extern_c_mod_items: Vec::new(),
-                bridge_items: Vec::new(),
-                bindgen_mod_items: Vec::new(),
-                materializations: Vec::new(),
                 global_items: vec![parse_quote! {
                     use super::#path;
                 }],
-                impl_entry: None,
                 extern_rust_mod_items: vec![parse_quote! {
                     type #id;
                 }],
+                ..Default::default()
             },
             Api::RustFn { sig, path, .. } => RsCodegenResult {
-                extern_c_mod_items: Vec::new(),
-                bridge_items: Vec::new(),
-                bindgen_mod_items: Vec::new(),
-                materializations: Vec::new(),
                 global_items: vec![parse_quote! {
                     use super::#path;
                 }],
-                impl_entry: None,
                 extern_rust_mod_items: vec![parse_quote! {
                     #sig;
                 }],
+                ..Default::default()
             },
             Api::RustSubclassFn {
                 details, subclass, ..
@@ -584,8 +564,8 @@ impl<'a> RsCodeGenerator<'a> {
                     subclasses_with_a_single_trivial_constructor.contains(&name.0.name);
                 self.generate_subclass(name, &superclass, methods, generate_peer_constructor)
             }
-            Api::RustSubclassConstructor { .. } => RsCodegenResult::default(),
             Api::IgnoredItem { err, ctx, .. } => Self::generate_error_entry(err, ctx),
+            Api::SubclassTraitItem { .. } => RsCodegenResult::default(),
         }
     }
 
@@ -634,11 +614,13 @@ impl<'a> RsCodeGenerator<'a> {
             let supers = SubclassName::get_supers_trait_name(superclass).to_type_path();
             let methods_impls: Vec<ImplItem> = methods
                 .iter()
+                .filter(|m| !m.is_pure_virtual)
                 .map(|m| {
-                    let supern = make_ident(format!("{}_super", m.name.to_string()));
+                    let cpp_super_method_name =
+                        SubclassName::get_super_fn_name(&Namespace::new(), &m.name.to_string())
+                            .get_final_ident();
                     let mut params = m.params.clone();
                     let ret = &m.ret_type.clone();
-                    let cpp_method_name = make_ident(format!("{}_super", m.name.to_string()));
                     let (peer_fn, first_param) = match m.receiver_mutability {
                         ReceiverMutability::Const => ("peer", parse_quote!(&self)),
                         ReceiverMutability::Mutable => ("peer_mut", parse_quote!(&mut self)),
@@ -646,21 +628,23 @@ impl<'a> RsCodeGenerator<'a> {
                     let peer_fn = make_ident(peer_fn);
                     *(params.iter_mut().next().unwrap()) = first_param;
                     let param_names = m.param_names.iter().skip(1);
-                    let unsafe_token = get_unsafe_token(m.requires_unsafe);
+                    let unsafe_token = m.requires_unsafe.wrapper_token();
                     parse_quote! {
-                        #unsafe_token fn #supern(#params) #ret {
+                        #unsafe_token fn #cpp_super_method_name(#params) #ret {
                             use autocxx::subclass::CppSubclass;
-                            self.#peer_fn().#cpp_method_name(#(#param_names),*)
+                            self.#peer_fn().#cpp_super_method_name(#(#param_names),*)
                         }
                     }
                 })
                 .collect();
-            bindgen_mod_items.push(parse_quote! {
-                #[allow(non_snake_case)]
-                impl #supers for super::super::super::#id {
-                    #(#methods_impls)*
-                }
-            });
+            if !methods_impls.is_empty() {
+                bindgen_mod_items.push(parse_quote! {
+                    #[allow(non_snake_case)]
+                    impl #supers for super::super::super::#id {
+                        #(#methods_impls)*
+                    }
+                });
+            }
         }
         if generate_peer_constructor {
             bindgen_mod_items.push(parse_quote! {
@@ -692,7 +676,7 @@ impl<'a> RsCodeGenerator<'a> {
         // TODO it would be nice to impl AsMut here but pin prevents us
         bindgen_mod_items.push(parse_quote! {
             impl super::super::super::#id {
-                pub fn pin_mut(&mut self) -> std::pin::Pin<&mut cxxbridge::#super_cxxxbridge_id> {
+                pub fn pin_mut(&mut self) -> ::std::pin::Pin<&mut cxxbridge::#super_cxxxbridge_id> {
                     use autocxx::subclass::CppSubclass;
                     self.peer_mut().#as_mut_id()
                 }
@@ -707,13 +691,15 @@ impl<'a> RsCodeGenerator<'a> {
         });
         RsCodegenResult {
             extern_c_mod_items,
-            bridge_items: create_impl_items(&cpp_id, self.config),
+            // For now we just assume we can't keep subclasses in vectors, but we can put them in
+            // smart pointers.
+            // That's the reason for the 'false' and 'true'
+            bridge_items: create_impl_items(&cpp_id, false, true, self.config),
             bindgen_mod_items,
             materializations: vec![Use::Custom(Box::new(parse_quote! {
                 pub use cxxbridge::#cpp_id;
             }))],
             global_items,
-            impl_entry: None,
             extern_rust_mod_items: vec![
                 parse_quote! {
                     pub type #holder;
@@ -722,6 +708,7 @@ impl<'a> RsCodeGenerator<'a> {
                     fn #remove_ownership(me: Box<#holder>) -> Box<#holder>;
                 },
             ],
+            ..Default::default()
         }
     }
 
@@ -732,7 +719,7 @@ impl<'a> RsCodeGenerator<'a> {
     ) -> RsCodegenResult {
         let params = details.params;
         let ret = details.ret;
-        let unsafe_token = get_unsafe_token(details.requires_unsafe);
+        let unsafe_token = details.requires_unsafe.wrapper_token();
         let global_def = quote! { #unsafe_token fn #api_name(#params) #ret };
         let params = unqualify_params(params);
         let ret = unqualify_ret_type(ret);
@@ -759,10 +746,6 @@ impl<'a> RsCodeGenerator<'a> {
         let destroy_panic_msg = format!("Rust subclass API (method {} of subclass {} of superclass {}) called after subclass destroyed", method_name, subclass.0.name, superclass_id);
         let reentrancy_panic_msg = format!("Rust subclass API (method {} of subclass {} of superclass {}) called whilst subclass already borrowed - likely a re-entrant call",  method_name, subclass.0.name, superclass_id);
         RsCodegenResult {
-            extern_c_mod_items: Vec::new(),
-            bridge_items: Vec::new(),
-            bindgen_mod_items: Vec::new(),
-            materializations: Vec::new(),
             global_items: vec![parse_quote! {
                 #global_def {
                     let rc = me.0
@@ -778,32 +761,32 @@ impl<'a> RsCodeGenerator<'a> {
                         #args)
                 }
             }],
-            impl_entry: None,
             extern_rust_mod_items: vec![ForeignItem::Fn(cxxbridge_decl)],
+            ..Default::default()
         }
     }
 
     fn args_from_sig(params: &Punctuated<FnArg, Comma>) -> impl Iterator<Item = Expr> + '_ {
-        params
-            .iter()
-            .skip(1)
-            .map(|fnarg| match fnarg {
-                syn::FnArg::Receiver(_) => None,
-                syn::FnArg::Typed(fnarg) => match &*fnarg.pat {
-                    syn::Pat::Ident(id) => Some(Self::id_to_expr(&id.ident)),
-                    _ => None,
-                },
-            })
-            .flatten()
+        params.iter().skip(1).filter_map(|fnarg| match fnarg {
+            syn::FnArg::Receiver(_) => None,
+            syn::FnArg::Typed(fnarg) => match &*fnarg.pat {
+                syn::Pat::Ident(id) => Some(Self::id_to_expr(&id.ident)),
+                _ => None,
+            },
+        })
     }
 
+    #[allow(clippy::too_many_arguments)] // currently the least unclear way
     fn generate_type<F>(
         &self,
         name: &QualifiedName,
         id: Ident,
         type_kind: TypeKind,
+        movable: bool,
+        destroyable: bool,
         item_creator: F,
         associated_methods: &HashMap<QualifiedName, Vec<SuperclassMethod>>,
+        layout: Option<Layout>,
     ) -> RsCodegenResult
     where
         F: FnOnce() -> Option<(Item, Option<Attribute>)>,
@@ -817,19 +800,28 @@ impl<'a> RsCodeGenerator<'a> {
             associated_methods.get(name),
         );
         let orig_item = item_creator();
+        // We have a choice here to either:
+        // a) tell cxx to generate an opaque type using 'type A;'
+        // b) generate a concrete type definition, e.g. by using bindgen's
+        //    or doing our own, and then telling cxx 'type A = bindgen::A';'
         match type_kind {
-            TypeKind::Pod | TypeKind::NonPodNested => {
+            TypeKind::Pod | TypeKind::NonPod => {
+                // Feed cxx "type T = root::bindgen::T"
+                // For non-POD types, there might be the option of simply giving
+                // cxx a "type T;" as we do for abstract types below. There's
+                // two reasons we don't:
+                // a) we want to specify size and alignment for the sake of
+                //    moveit;
+                // b) for nested types such as 'A::B', there is no combination
+                //    of cxx-acceptable attributes which will inform cxx that
+                //    A is a class rather than a namespace.
                 let mut item = orig_item
                     .expect("Instantiable types must provide instance")
                     .0;
-                if matches!(type_kind, TypeKind::NonPodNested) {
-                    // We have to use 'type A = super::bindgen::A::B'
-                    // because if we use simply 'type A', there is no combination
-                    // of cxx-acceptable attributes which will inform cxx that
-                    // A is a class rather than a namespace.
+                if matches!(type_kind, TypeKind::NonPod) {
                     if let Item::Struct(ref mut s) = item {
                         // Retain generics and doc attrs.
-                        make_non_pod(s);
+                        make_non_pod(s, layout);
                     } else {
                         // enum
                         item = Item::Struct(new_non_pod_struct(id.clone()));
@@ -838,25 +830,24 @@ impl<'a> RsCodeGenerator<'a> {
                 bindgen_mod_items.push(item);
                 RsCodegenResult {
                     global_items: self.generate_extern_type_impl(type_kind, name),
-                    impl_entry: None,
-                    bridge_items: create_impl_items(&id, self.config),
+                    bridge_items: create_impl_items(&id, movable, destroyable, self.config),
                     extern_c_mod_items: vec![self.generate_cxxbridge_type(name, true, None)],
                     bindgen_mod_items,
                     materializations,
-                    extern_rust_mod_items: Vec::new(),
+                    ..Default::default()
                 }
             }
-            TypeKind::NonPod | TypeKind::Abstract => {
+            TypeKind::Abstract => {
+                // Feed cxx "type T;"
+                // We MUST do this because otherwise cxx assumes this can be
+                // instantiated using UniquePtr etc.
                 bindgen_mod_items.push(Item::Use(parse_quote! { pub use cxxbridge::#id; }));
-                let doc_attr = orig_item.map(|maybe_item| maybe_item.1).flatten();
+                let doc_attr = orig_item.and_then(|maybe_item| maybe_item.1);
                 RsCodegenResult {
                     extern_c_mod_items: vec![self.generate_cxxbridge_type(name, false, doc_attr)],
-                    extern_rust_mod_items: Vec::new(),
-                    bridge_items: Vec::new(),
-                    global_items: Vec::new(),
                     bindgen_mod_items,
-                    impl_entry: None,
                     materializations,
+                    ..Default::default()
                 }
             }
         }
@@ -873,7 +864,9 @@ impl<'a> RsCodeGenerator<'a> {
                 .iter()
                 .map(|method| {
                     let id = &method.name;
-                    let super_id = make_ident(format!("{}_super", id.to_string()));
+                    let super_id =
+                        SubclassName::get_super_fn_name(&Namespace::new(), &id.to_string())
+                            .get_final_ident();
                     let param_names: Punctuated<Expr, Comma> =
                         Self::args_from_sig(&method.params).collect();
                     let mut params = method.params.clone();
@@ -882,36 +875,53 @@ impl<'a> RsCodeGenerator<'a> {
                         ReceiverMutability::Mutable => parse_quote!(&mut self),
                     };
                     let ret_type = &method.ret_type;
-                    let unsafe_token = get_unsafe_token(method.requires_unsafe);
-                    let a: TraitItem = parse_quote!(
-                        #unsafe_token fn #super_id(#params) #ret_type;
-                    );
-                    let b: TraitItem = parse_quote!(
-                        #unsafe_token fn #id(#params) #ret_type {
-                            self.#super_id(#param_names)
-                        }
-                    );
-                    (a, b)
+                    let unsafe_token = method.requires_unsafe.wrapper_token();
+                    if method.is_pure_virtual {
+                        (
+                            None,
+                            parse_quote!(
+                                #unsafe_token fn #id(#params) #ret_type;
+                            ),
+                        )
+                    } else {
+                        let a: Option<TraitItem> = Some(parse_quote!(
+                            #unsafe_token fn #super_id(#params) #ret_type;
+                        ));
+                        let b: TraitItem = parse_quote!(
+                            #unsafe_token fn #id(#params) #ret_type {
+                                self.#super_id(#param_names)
+                            }
+                        );
+                        (a, b)
+                    }
                 })
                 .unzip();
+            let supers: Vec<_> = supers.into_iter().flatten().collect();
             let supers_name = SubclassName::get_supers_trait_name(name).get_final_ident();
             let methods_name = SubclassName::get_methods_trait_name(name).get_final_ident();
-            bindgen_mod_items.push(parse_quote! {
-                #[allow(non_snake_case)]
-                pub trait #supers_name {
-                    #(#supers)*
-                }
-            });
-            bindgen_mod_items.push(parse_quote! {
-                #[allow(non_snake_case)]
-                pub trait #methods_name : #supers_name {
-                    #(#mains)*
-                }
-            });
-            materializations.extend([
-                Use::SpecificNameFromBindgen(methods_name),
-                Use::SpecificNameFromBindgen(supers_name),
-            ]);
+            if !supers.is_empty() {
+                bindgen_mod_items.push(parse_quote! {
+                    #[allow(non_snake_case)]
+                    pub trait #supers_name {
+                        #(#supers)*
+                    }
+                });
+                bindgen_mod_items.push(parse_quote! {
+                    #[allow(non_snake_case)]
+                    pub trait #methods_name : #supers_name {
+                        #(#mains)*
+                    }
+                });
+                materializations.push(Use::SpecificNameFromBindgen(supers_name));
+            } else {
+                bindgen_mod_items.push(parse_quote! {
+                    #[allow(non_snake_case)]
+                    pub trait #methods_name {
+                        #(#mains)*
+                    }
+                });
+            }
+            materializations.push(Use::SpecificNameFromBindgen(methods_name));
         }
     }
 
@@ -952,11 +962,7 @@ impl<'a> RsCodeGenerator<'a> {
             }
             ErrorContext::Method { self_ty, method } => {
                 // If the type can't be represented (e.g. u8) this would get fiddly.
-                let id = make_ident(format!(
-                    "{}_method_{}",
-                    self_ty.to_string(),
-                    method.to_string()
-                ));
+                let id = make_ident(format!("{}_method_{}", self_ty, method));
                 (
                     None,
                     Some(Use::Custom(Box::new(parse_quote! {
@@ -965,15 +971,12 @@ impl<'a> RsCodeGenerator<'a> {
                     }))),
                 )
             }
+            ErrorContext::NoCode => (None, None),
         };
         RsCodegenResult {
-            global_items: Vec::new(),
             impl_entry,
-            bridge_items: Vec::new(),
-            extern_c_mod_items: Vec::new(),
-            bindgen_mod_items: Vec::new(),
             materializations: materialization.into_iter().collect(),
-            extern_rust_mod_items: Vec::new(),
+            ..Default::default()
         }
     }
 
@@ -1090,27 +1093,19 @@ impl<'a> RsCodeGenerator<'a> {
     }
 }
 
-fn get_unsafe_token(requires_unsafe: bool) -> TokenStream {
-    if requires_unsafe {
-        quote! { unsafe }
-    } else {
-        quote! {}
-    }
-}
-
-fn find_trivially_constructed_subclasses(apis: &[Api<FnPhase>]) -> HashSet<QualifiedName> {
+fn find_trivially_constructed_subclasses(apis: &ApiVec<FnPhase>) -> HashSet<QualifiedName> {
     let (simple_constructors, complex_constructors): (Vec<_>, Vec<_>) = apis
         .iter()
-        .map(|api| match api {
-            Api::RustSubclassConstructor {
-                subclass,
-                is_trivial,
-                ..
-            } => Some((&subclass.0.name, is_trivial)),
+        .filter_map(|api| match api {
+            Api::Function { fun, .. } => match &fun.provenance {
+                Provenance::SynthesizedSubclassConstructor(details) => {
+                    Some((&details.subclass.0.name, details.is_trivial))
+                }
+                _ => None,
+            },
             _ => None,
         })
-        .flatten()
-        .partition(|(_, trivial)| **trivial);
+        .partition(|(_, trivial)| *trivial);
     let simple_constructors: HashSet<_> =
         simple_constructors.into_iter().map(|(qn, _)| qn).collect();
     let complex_constructors: HashSet<_> =
@@ -1143,5 +1138,6 @@ struct RsCodegenResult {
     global_items: Vec<Item>,
     bindgen_mod_items: Vec<Item>,
     impl_entry: Option<Box<ImplBlockDetails>>,
+    trait_impl_entry: Option<Box<TraitImplBlockDetails>>,
     materializations: Vec<Use>,
 }
