@@ -1,19 +1,15 @@
 // Copyright 2020 Google LLC
 //
-// Licensed under the Apache License, Version 2.0 (the "License");
-// you may not use this file except in compliance with the License.
-// You may obtain a copy of the License at
-//
-//    https://www.apache.org/licenses/LICENSE-2.0
-//
-// Unless required by applicable law or agreed to in writing, software
-// distributed under the License is distributed on an "AS IS" BASIS,
-// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-// See the License for the specific language governing permissions and
-// limitations under the License.
+// Licensed under the Apache License, Version 2.0 <LICENSE-APACHE or
+// https://www.apache.org/licenses/LICENSE-2.0> or the MIT license
+// <LICENSE-MIT or https://opensource.org/licenses/MIT>, at your
+// option. This file may not be copied, modified, or distributed
+// except according to those terms.
 
+use std::collections::HashSet;
 use std::fmt::Display;
 
+use itertools::Itertools;
 use syn::Ident;
 
 use crate::types::{Namespace, QualifiedName};
@@ -28,7 +24,6 @@ pub enum ConvertError {
     ComplexTypedefTarget(String),
     UnexpectedThisType(Namespace, String),
     UnsupportedBuiltInType(QualifiedName),
-    VirtualThisType(Namespace, String),
     ConflictingTemplatedArgsWithTypedef(QualifiedName),
     UnacceptableParam(String),
     NotOneInputReference(String),
@@ -45,14 +40,23 @@ pub enum ConvertError {
     UnusedTemplateParam,
     TooManyUnderscores,
     UnknownDependentType(QualifiedName),
-    IgnoredDependent,
-    MoveConstructorUnsupported,
-    ReservedName,
+    IgnoredDependent(HashSet<QualifiedName>),
+    ReservedName(String),
     DuplicateCxxBridgeName,
     UnsupportedReceiver,
     BoxContainingNonRustType(QualifiedName),
     RustTypeWithAPath(QualifiedName),
     AbstractNestedType,
+    NonPublicNestedType,
+    RValueParam,
+    RValueReturn,
+    PrivateMethod,
+    AssignmentOperator,
+    Deleted,
+    RValueReferenceField,
+    MethodOfNonAllowlistedType,
+    MethodOfGenericType,
+    DuplicateItemsFoundInParsing,
 }
 
 fn format_maybe_identifier(id: &Option<Ident>) -> String {
@@ -73,7 +77,6 @@ impl Display for ConvertError {
             ConvertError::ComplexTypedefTarget(ty) => write!(f, "autocxx was unable to produce a typdef pointing to the complex type {}.", ty)?,
             ConvertError::UnexpectedThisType(ns, fn_name) => write!(f, "Unexpected type for 'this' in the function {}{}.", fn_name, ns.to_display_suffix())?,
             ConvertError::UnsupportedBuiltInType(ty) => write!(f, "autocxx does not yet know how to support the built-in C++ type {} - please raise an issue on github", ty.to_cpp_name())?,
-            ConvertError::VirtualThisType(ns, fn_name) => write!(f, "Member function encountered where the 'this' type is 'void*', but we were unable to recognize which type that corresponds to. Function {}{}.", fn_name, ns.to_display_suffix())?,
             ConvertError::ConflictingTemplatedArgsWithTypedef(tn) => write!(f, "Type {} has templated arguments and so does the typedef to which it points", tn)?,
             ConvertError::UnacceptableParam(fn_name) => write!(f, "Function {} has a parameter or return type which is either on the blocklist or a forward declaration", fn_name)?,
             ConvertError::NotOneInputReference(fn_name) => write!(f, "Function {} has a return reference parameter, but 0 or >1 input reference parameters, so the lifetime of the output reference cannot be deduced.", fn_name)?,
@@ -90,14 +93,23 @@ impl Display for ConvertError {
             ConvertError::UnusedTemplateParam => write!(f, "This function or method uses a type where one of the template parameters was incomprehensible to bindgen/autocxx - probably because it uses template specialization.")?,
             ConvertError::TooManyUnderscores => write!(f, "Names containing __ are reserved by C++ so not acceptable to cxx")?,
             ConvertError::UnknownDependentType(qn) => write!(f, "This item relies on a type not known to autocxx ({})", qn.to_cpp_name())?,
-            ConvertError::IgnoredDependent => write!(f, "This item depends on some other type which autocxx could not generate.")?,
-            ConvertError::MoveConstructorUnsupported => write!(f, "This is a move constructor, for which we currently cannot generate bindings.")?,
-            ConvertError::ReservedName => write!(f, "The item name is a reserved word in Rust.")?,
+            ConvertError::IgnoredDependent(qns) => write!(f, "This item depends on some other type(s) which autocxx could not generate, some of them are: {}", qns.iter().join(", "))?,
+            ConvertError::ReservedName(id) => write!(f, "The item name '{}' is a reserved word in Rust.", id)?,
             ConvertError::DuplicateCxxBridgeName => write!(f, "This item name is used in multiple namespaces. At present, autocxx and cxx allow only one type of a given name. This limitation will be fixed in future.")?,
             ConvertError::UnsupportedReceiver => write!(f, "This is a method on a type which can't be used as the receiver in Rust (i.e. self/this). This is probably because some type involves template specialization.")?,
             ConvertError::BoxContainingNonRustType(ty) => write!(f, "A rust::Box<T> was encountered where T was not known to be a Rust type. Use rust_type!(T): {}", ty.to_cpp_name())?,
             ConvertError::RustTypeWithAPath(ty) => write!(f, "A qualified Rust type was found (i.e. one containing ::): {}. Rust types must always be a simple identifier.", ty.to_cpp_name())?,
             ConvertError::AbstractNestedType => write!(f, "This type is nested within another struct/class, yet is abstract (or is not on the allowlist so we can't be sure). This is not yet supported by autocxx. If you don't believe this type is abstract, add it to the allowlist.")?,
+            ConvertError::NonPublicNestedType => write!(f, "This type is nested within another struct/class with protected or private visibility.")?,
+            ConvertError::RValueParam => write!(f, "This function takes an rvalue reference parameter (&&) which is not yet supported.")?,
+            ConvertError::RValueReturn => write!(f, "This function returns an rvalue reference (&&) which is not yet supported.")?,
+            ConvertError::PrivateMethod => write!(f, "This method is private")?,
+            ConvertError::AssignmentOperator => write!(f, "autocxx does not know how to generate bindings to operator=")?,
+            ConvertError::Deleted => write!(f, "This function was marked =delete")?,
+            ConvertError::RValueReferenceField => write!(f, "This structure has an rvalue reference field (&&) which is not yet supported.")?,
+            ConvertError::MethodOfNonAllowlistedType => write!(f, "This type was not on the allowlist, so we are not generating methods for it.")?,
+            ConvertError::MethodOfGenericType => write!(f, "This type is templated, so we can't generate bindings. We will instead generate bindings for each instantiation.")?,
+            ConvertError::DuplicateItemsFoundInParsing => write!(f, "bindgen generated multiple different APIs (functions/types) with this name. autocxx doesn't know how to diambiguate them, so we won't generate bindings for any of them.")?,
         }
         Ok(())
     }
@@ -106,15 +118,23 @@ impl Display for ConvertError {
 #[derive(Clone)]
 pub(crate) enum ErrorContext {
     Item(Ident),
-    Method { self_ty: Ident, method: Ident },
+    Method {
+        self_ty: Ident,
+        method: Ident,
+    },
+    /// This error means the context where the code would be generated doesn't exist, so there's
+    /// nowhere to attach the error, but we still want to record that we had an error generating
+    /// it.
+    NoCode,
 }
 
 impl ErrorContext {
     /// Return the ID in the output mod with which this should be associated
-    pub(crate) fn get_id(&self) -> &Ident {
+    pub(crate) fn get_id(&self) -> Option<&Ident> {
         match self {
-            ErrorContext::Item(id) => id,
-            ErrorContext::Method { self_ty, method: _ } => self_ty,
+            ErrorContext::Item(id) => Some(id),
+            ErrorContext::Method { self_ty, method: _ } => Some(self_ty),
+            ErrorContext::NoCode => None,
         }
     }
 }
@@ -124,10 +144,12 @@ impl std::fmt::Display for ErrorContext {
         match self {
             ErrorContext::Item(id) => write!(f, "{}", id),
             ErrorContext::Method { self_ty, method } => write!(f, "{}::{}", self_ty, method),
+            ErrorContext::NoCode => write!(f, "<not generated>"),
         }
     }
 }
 
+#[derive(Clone)]
 pub(crate) struct ConvertErrorWithContext(pub(crate) ConvertError, pub(crate) Option<ErrorContext>);
 
 impl std::fmt::Debug for ConvertErrorWithContext {
