@@ -1,22 +1,17 @@
 // Copyright 2020 Google LLC
 //
-// Licensed under the Apache License, Version 2.0 (the "License");
-// you may not use this file except in compliance with the License.
-// You may obtain a copy of the License at
-//
-//    https://www.apache.org/licenses/LICENSE-2.0
-//
-// Unless required by applicable law or agreed to in writing, software
-// distributed under the License is distributed on an "AS IS" BASIS,
-// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-// See the License for the specific language governing permissions and
-// limitations under the License.
+// Licensed under the Apache License, Version 2.0 <LICENSE-APACHE or
+// https://www.apache.org/licenses/LICENSE-2.0> or the MIT license
+// <LICENSE-MIT or https://opensource.org/licenses/MIT>, at your
+// option. This file may not be copied, modified, or distributed
+// except according to those terms.
 
 use std::collections::HashSet;
 
 use crate::{
     conversion::{
-        api::{Api, ApiName, SubclassName, TypedefKind, UnanalyzedApi},
+        api::{Api, ApiName, NullPhase, StructDetails, SubclassName, TypedefKind, UnanalyzedApi},
+        apivec::ApiVec,
         ConvertError,
     },
     types::Namespace,
@@ -30,34 +25,28 @@ use crate::{
     types::validate_ident_ok_for_cxx,
 };
 use autocxx_parser::IncludeCppConfig;
-use syn::{parse_quote, Attribute, Fields, Ident, Item, LitStr, TypePath, UseTree};
+use syn::{parse_quote, Fields, Ident, Item, TypePath, UseTree};
 
-use super::super::utilities::generate_utilities;
+use super::{
+    super::utilities::generate_utilities, bindgen_semantic_attributes::BindgenSemanticAttributes,
+};
 
 use super::parse_foreign_mod::ParseForeignMod;
 
 /// Parses a bindgen mod in order to understand the APIs within it.
 pub(crate) struct ParseBindgen<'a> {
     config: &'a IncludeCppConfig,
-    apis: Vec<UnanalyzedApi>,
-    /// Here we track the last struct which bindgen told us about.
-    /// Any subsequent "extern 'C'" blocks are methods belonging to that type,
-    /// even if the 'this' is actually recorded as void in the
-    /// function signature.
-    latest_virtual_this_type: Option<QualifiedName>,
+    apis: ApiVec<NullPhase>,
 }
 
-fn api_name(ns: &Namespace, id: Ident, attrs: &[Attribute]) -> ApiName {
-    ApiName {
-        name: QualifiedName::new(ns, id),
-        cpp_name: get_bindgen_original_name_annotation(attrs),
-    }
+fn api_name(ns: &Namespace, id: Ident, attrs: &BindgenSemanticAttributes) -> ApiName {
+    ApiName::new_with_cpp_name(ns, id, attrs.get_original_name())
 }
 
 pub(crate) fn api_name_qualified(
     ns: &Namespace,
     id: Ident,
-    attrs: &[Attribute],
+    attrs: &BindgenSemanticAttributes,
 ) -> Result<ApiName, ConvertErrorWithContext> {
     match validate_ident_ok_for_cxx(&id.to_string()) {
         Err(e) => {
@@ -68,29 +57,11 @@ pub(crate) fn api_name_qualified(
     }
 }
 
-pub(super) fn get_bindgen_original_name_annotation(attrs: &[Attribute]) -> Option<String> {
-    attrs
-        .iter()
-        .filter_map(|a| {
-            if a.path.is_ident("bindgen_original_name") {
-                let r: Result<LitStr, syn::Error> = a.parse_args();
-                match r {
-                    Ok(ls) => Some(ls.value()),
-                    Err(_) => None,
-                }
-            } else {
-                None
-            }
-        })
-        .next()
-}
-
 impl<'a> ParseBindgen<'a> {
     pub(crate) fn new(config: &'a IncludeCppConfig) -> Self {
         ParseBindgen {
             config,
-            apis: Vec::new(),
-            latest_virtual_this_type: None,
+            apis: ApiVec::new(),
         }
     }
 
@@ -99,7 +70,7 @@ impl<'a> ParseBindgen<'a> {
     pub(crate) fn parse_items(
         mut self,
         items: Vec<Item>,
-    ) -> Result<Vec<UnanalyzedApi>, ConvertError> {
+    ) -> Result<ApiVec<NullPhase>, ConvertError> {
         let items = Self::find_items_in_root(items)?;
         if !self.config.exclude_utilities() {
             generate_utilities(&mut self.apis, self.config);
@@ -161,7 +132,7 @@ impl<'a> ParseBindgen<'a> {
         // This object maintains some state specific to this namespace, i.e.
         // this particular mod.
         let mut mod_converter = ParseForeignMod::new(ns.clone());
-        let mut more_apis = Vec::new();
+        let mut more_apis = ApiVec::new();
         for item in items {
             report_any_error(&ns, &mut more_apis, || {
                 self.parse_item(item, &mut mod_converter, &ns)
@@ -179,8 +150,7 @@ impl<'a> ParseBindgen<'a> {
     ) -> Result<(), ConvertErrorWithContext> {
         match item {
             Item::ForeignMod(fm) => {
-                mod_converter
-                    .convert_foreign_mod_items(fm.items, self.latest_virtual_this_type.clone());
+                mod_converter.convert_foreign_mod_items(fm.items);
                 Ok(())
             }
             Item::Struct(s) => {
@@ -188,35 +158,44 @@ impl<'a> ParseBindgen<'a> {
                     return Ok(());
                 }
                 let is_forward_declaration = Self::spot_forward_declaration(&s.fields);
+                let annotations = BindgenSemanticAttributes::new(&s.attrs);
                 // cxx::bridge can't cope with type aliases to generic
                 // types at the moment.
-                let name = api_name_qualified(ns, s.ident.clone(), &s.attrs)?;
+                let name = api_name_qualified(ns, s.ident.clone(), &annotations)?;
                 let api = if ns.is_empty() && self.config.is_rust_type(&s.ident) {
                     None
                 } else if is_forward_declaration {
                     Some(UnanalyzedApi::ForwardDeclaration { name })
                 } else {
+                    let has_rvalue_reference_fields = s.fields.iter().any(|f| {
+                        BindgenSemanticAttributes::new(&f.attrs).has_attr("rvalue_reference")
+                    });
                     Some(UnanalyzedApi::Struct {
                         name,
-                        item: s,
+                        details: Box::new(StructDetails {
+                            vis: annotations.get_cpp_visibility(),
+                            layout: annotations.get_layout(),
+                            item: s,
+                            has_rvalue_reference_fields,
+                        }),
                         analysis: (),
                     })
                 };
                 if let Some(api) = api {
-                    self.latest_virtual_this_type = Some(api.name().clone());
                     if !self.config.is_on_blocklist(&api.name().to_cpp_name()) {
-                        self.apis.push(api);
+                        self.apis.push_eliminating_duplicates(api);
                     }
                 }
                 Ok(())
             }
             Item::Enum(e) => {
+                let annotations = BindgenSemanticAttributes::new(&e.attrs);
                 let api = UnanalyzedApi::Enum {
-                    name: api_name_qualified(ns, e.ident.clone(), &e.attrs)?,
+                    name: api_name_qualified(ns, e.ident.clone(), &annotations)?,
                     item: e,
                 };
                 if !self.config.is_on_blocklist(&api.name().to_cpp_name()) {
-                    self.apis.push(api);
+                    self.apis.push_eliminating_duplicates(api);
                 }
                 Ok(())
             }
@@ -272,14 +251,16 @@ impl<'a> ParseBindgen<'a> {
                                     Some(ErrorContext::Item(new_id.clone())),
                                 ));
                             }
-                            self.apis.push(UnanalyzedApi::Typedef {
-                                name: api_name(ns, new_id.clone(), &use_item.attrs),
-                                item: TypedefKind::Use(parse_quote! {
-                                    pub use #old_path as #new_id;
-                                }),
-                                old_tyname: Some(old_tyname),
-                                analysis: (),
-                            });
+                            let annotations = BindgenSemanticAttributes::new(&use_item.attrs);
+                            self.apis
+                                .push_eliminating_duplicates(UnanalyzedApi::Typedef {
+                                    name: api_name(ns, new_id.clone(), &annotations),
+                                    item: TypedefKind::Use(parse_quote! {
+                                        pub use #old_path as #new_id;
+                                    }),
+                                    old_tyname: Some(old_tyname),
+                                    analysis: (),
+                                });
                             break;
                         }
                         _ => {
@@ -293,19 +274,24 @@ impl<'a> ParseBindgen<'a> {
                 Ok(())
             }
             Item::Const(const_item) => {
-                self.apis.push(UnanalyzedApi::Const {
-                    name: api_name(ns, const_item.ident.clone(), &const_item.attrs),
+                let annotations = BindgenSemanticAttributes::new(&const_item.attrs);
+                self.apis.push_eliminating_duplicates(UnanalyzedApi::Const {
+                    name: api_name(ns, const_item.ident.clone(), &annotations),
                     const_item,
                 });
                 Ok(())
             }
             Item::Type(ity) => {
-                self.apis.push(UnanalyzedApi::Typedef {
-                    name: api_name(ns, ity.ident.clone(), &ity.attrs),
-                    item: TypedefKind::Type(ity),
-                    old_tyname: None,
-                    analysis: (),
-                });
+                let annotations = BindgenSemanticAttributes::new(&ity.attrs);
+                // It's known that sometimes bindgen will give us duplicate typedefs with the
+                // same name - see test_issue_264.
+                self.apis
+                    .push_eliminating_duplicates(UnanalyzedApi::Typedef {
+                        name: api_name(ns, ity.ident.clone(), &annotations),
+                        item: TypedefKind::Type(ity),
+                        old_tyname: None,
+                        analysis: (),
+                    });
                 Ok(())
             }
             _ => Err(ConvertErrorWithContext(
