@@ -1,18 +1,13 @@
 // Copyright 2020 Google LLC
 //
-// Licensed under the Apache License, Version 2.0 (the "License");
-// you may not use this file except in compliance with the License.
-// You may obtain a copy of the License at
-//
-//    https://www.apache.org/licenses/LICENSE-2.0
-//
-// Unless required by applicable law or agreed to in writing, software
-// distributed under the License is distributed on an "AS IS" BASIS,
-// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-// See the License for the specific language governing permissions and
-// limitations under the License.
+// Licensed under the Apache License, Version 2.0 <LICENSE-APACHE or
+// https://www.apache.org/licenses/LICENSE-2.0> or the MIT license
+// <LICENSE-MIT or https://opensource.org/licenses/MIT>, at your
+// option. This file may not be copied, modified, or distributed
+// except according to those terms.
 
-use crate::conversion::api::{ApiName, CppVisibility};
+use crate::conversion::api::{ApiName, NullPhase, Provenance};
+use crate::conversion::apivec::ApiVec;
 use crate::conversion::doc_attr::get_doc_attr;
 use crate::conversion::error_reporter::report_any_error;
 use crate::conversion::{
@@ -24,13 +19,10 @@ use crate::{
     conversion::ConvertError,
     types::{Namespace, QualifiedName},
 };
-use std::collections::{HashMap, HashSet};
-use syn::{
-    Block, Expr, ExprCall, ForeignItem, ForeignItemFn, Ident, ImplItem, ItemImpl, LitStr, Stmt,
-    Type,
-};
+use std::collections::HashMap;
+use syn::{Block, Expr, ExprCall, ForeignItem, Ident, ImplItem, ItemImpl, Stmt, Type};
 
-use super::parse_bindgen::get_bindgen_original_name_annotation;
+use super::bindgen_semantic_attributes::BindgenSemanticAttributes;
 
 /// Parses a given bindgen-generated 'mod' into suitable
 /// [Api]s. In bindgen output, a given mod concerns
@@ -47,7 +39,7 @@ pub(crate) struct ParseForeignMod {
     // may actually be methods (static or otherwise). Mapping from
     // function name to type name.
     method_receivers: HashMap<Ident, QualifiedName>,
-    ignored_apis: Vec<UnanalyzedApi>,
+    ignored_apis: ApiVec<NullPhase>,
 }
 
 impl ParseForeignMod {
@@ -56,67 +48,46 @@ impl ParseForeignMod {
             ns,
             funcs_to_convert: Vec::new(),
             method_receivers: HashMap::new(),
-            ignored_apis: Vec::new(),
+            ignored_apis: ApiVec::new(),
         }
     }
 
     /// Record information from foreign mod items encountered
     /// in bindgen output.
-    pub(crate) fn convert_foreign_mod_items(
-        &mut self,
-        foreign_mod_items: Vec<ForeignItem>,
-        virtual_this_type: Option<QualifiedName>,
-    ) {
-        let mut extra_apis = Vec::new();
+    pub(crate) fn convert_foreign_mod_items(&mut self, foreign_mod_items: Vec<ForeignItem>) {
+        let mut extra_apis = ApiVec::new();
         for i in foreign_mod_items {
             report_any_error(&self.ns.clone(), &mut extra_apis, || {
-                self.parse_foreign_item(i, &virtual_this_type)
+                self.parse_foreign_item(i)
             });
         }
         self.ignored_apis.append(&mut extra_apis);
     }
 
-    fn get_cpp_visibility(item: &ForeignItemFn) -> CppVisibility {
-        if Self::has_attr(&item, "bindgen_visibility_private") {
-            CppVisibility::Private
-        } else if Self::has_attr(&item, "bindgen_visibility_protected") {
-            CppVisibility::Protected
-        } else {
-            CppVisibility::Public
-        }
-    }
-
-    fn parse_foreign_item(
-        &mut self,
-        i: ForeignItem,
-        virtual_this_type: &Option<QualifiedName>,
-    ) -> Result<(), ConvertErrorWithContext> {
+    fn parse_foreign_item(&mut self, i: ForeignItem) -> Result<(), ConvertErrorWithContext> {
         match i {
             ForeignItem::Fn(item) => {
-                let cpp_vis = Self::get_cpp_visibility(&item);
-                let is_pure_virtual = Self::has_attr(&item, "bindgen_pure_virtual");
-                let unused_template_param =
-                    Self::has_attr(&item, "bindgen_unused_template_param_in_arg_or_return");
-                let is_move_constructor = Self::is_move_constructor(&item);
-                let (reference_args, return_type_is_reference) =
-                    Self::get_reference_parameters_and_return(&item);
-                let original_name = get_bindgen_original_name_annotation(&item.attrs);
+                let annotations = BindgenSemanticAttributes::new(&item.attrs);
                 let doc_attr = get_doc_attr(&item.attrs);
                 self.funcs_to_convert.push(FuncToConvert {
-                    virtual_this_type: virtual_this_type.clone(),
+                    provenance: Provenance::Bindgen,
                     self_ty: None,
                     ident: item.sig.ident,
                     doc_attr,
                     inputs: item.sig.inputs,
                     output: item.sig.output,
                     vis: item.vis,
-                    is_pure_virtual,
-                    cpp_vis,
-                    is_move_constructor,
-                    unused_template_param,
-                    return_type_is_reference,
-                    reference_args,
-                    original_name,
+                    virtualness: annotations.get_virtualness(),
+                    cpp_vis: annotations.get_cpp_visibility(),
+                    special_member: annotations.special_member_kind(),
+                    unused_template_param: annotations
+                        .has_attr("unused_template_param_in_arg_or_return"),
+                    references: annotations.get_reference_parameters_and_return(),
+                    original_name: annotations.get_original_name(),
+                    synthesized_this_type: None,
+                    add_to_trait: None,
+                    is_deleted: annotations.has_attr("deleted"),
+                    synthetic_cpp: None,
                 });
                 Ok(())
             }
@@ -131,47 +102,6 @@ impl ParseForeignMod {
         }
     }
 
-    fn has_attr(fun: &ForeignItemFn, attr_name: &str) -> bool {
-        fun.attrs.iter().any(|a| a.path.is_ident(attr_name))
-    }
-
-    fn get_bindgen_special_member_annotation(fun: &ForeignItemFn) -> Option<String> {
-        fun.attrs
-            .iter()
-            .filter_map(|a| {
-                if a.path.is_ident("bindgen_special_member") {
-                    let r: Result<LitStr, syn::Error> = a.parse_args();
-                    match r {
-                        Ok(ls) => Some(ls.value()),
-                        Err(_) => None,
-                    }
-                } else {
-                    None
-                }
-            })
-            .next()
-    }
-
-    fn is_move_constructor(fun: &ForeignItemFn) -> bool {
-        Self::get_bindgen_special_member_annotation(fun).map_or(false, |val| val == "move_ctor")
-    }
-
-    fn get_reference_parameters_and_return(fun: &ForeignItemFn) -> (HashSet<Ident>, bool) {
-        let mut ref_params = HashSet::new();
-        let mut ref_return = false;
-        for a in &fun.attrs {
-            if a.path.is_ident("bindgen_ret_type_reference") {
-                ref_return = true;
-            } else if a.path.is_ident("bindgen_arg_type_reference") {
-                let r: Result<Ident, syn::Error> = a.parse_args();
-                if let Ok(ls) = r {
-                    ref_params.insert(ls);
-                }
-            }
-        }
-        (ref_params, ref_return)
-    }
-
     /// Record information from impl blocks encountered in bindgen
     /// output.
     pub(crate) fn convert_impl_items(&mut self, imp: ItemImpl) {
@@ -181,13 +111,9 @@ impl ParseForeignMod {
         };
         for i in imp.items {
             if let ImplItem::Method(itm) = i {
-                let effective_fun_name = if itm.sig.ident == "new" {
-                    ty_id.clone()
-                } else {
-                    match get_called_function(&itm.block) {
-                        Some(id) => id.clone(),
-                        None => itm.sig.ident,
-                    }
+                let effective_fun_name = match get_called_function(&itm.block) {
+                    Some(id) => id.clone(),
+                    None => itm.sig.ident,
                 };
                 self.method_receivers.insert(
                     effective_fun_name,
@@ -200,12 +126,12 @@ impl ParseForeignMod {
     /// Indicate that all foreign mods and all impl blocks have been
     /// fed into us, and we should process that information to generate
     /// the resulting APIs.
-    pub(crate) fn finished(mut self, apis: &mut Vec<UnanalyzedApi>) {
+    pub(crate) fn finished(mut self, apis: &mut ApiVec<NullPhase>) {
         apis.append(&mut self.ignored_apis);
         while !self.funcs_to_convert.is_empty() {
             let mut fun = self.funcs_to_convert.remove(0);
             fun.self_ty = self.method_receivers.get(&fun.ident).cloned();
-            apis.push(UnanalyzedApi::Function {
+            apis.push_eliminating_duplicates(UnanalyzedApi::Function {
                 name: ApiName::new_with_cpp_name(
                     &self.ns,
                     fun.ident.clone(),
